<?php

/**
 * Convert MOD files to NinjaTracker+ format.
 * 
 * by Jesse Blue / Ninjaforce
 *
 * 
 * July 29, 2018 > Initial release
 * Nov  25, 2019 > Bug fixes:
 *                 - MODs with 1/2/3 or 15 channels were not converted correctly
 *                 - The converter now rotates loops near a zero-crossing to reduce the swap bug
 * Feb   8, 2020 > Bug fixes:
 *                 - some error messages were not displayed properly
 * Apr   3, 2021 > Bug fixes:
 *                 - Octaves 0 and 4 had wrong values
 * Jan  02, 2022 > Added support to stream samples, allowing to play modules with large samples
 *                 - File format version 1
 * Jan  16, 2022 > Added support for one-oscillator-tracks
 *                 - Song name and instrument names are now included in NTP files
 *                 - File format version 2
 **/

/* the converter class, can be used on its own */

class modinstrument
{
    public const TYPE_SINGLE = 0;
    public const TYPE_LOOPED = 1;
    public const TYPE_LOOPWITHHEADER = 2;

    public const STOPPER_BYTES_COUNT = 8;

    public $number = 0;
    public $name = '';
    public $volume = 0;
    public $finetune = 0;
    public $repeat_pointer = 0;
    public $repeat_length = 0;
    public $length = 0;
    public $data = [];

    private $must_stream = false;

    public function get_instrument_type()
    {
        if ($this->repeat_length < 3) {
            /* instrument is not looped */
            $type = self::TYPE_SINGLE;
        } elseif ($this->repeat_pointer > 2) {
            /* instrument is looped with a header */
            $type = self::TYPE_LOOPWITHHEADER;
        } else {
            /* instrument consists only of a loop */
            $type = self::TYPE_LOOPED;
        }

        return $type;
    }

    public function must_be_splitted(): bool
    {
        return ($this->repeat_length > 2 && $this->repeat_pointer > 2);
    }

    public function get_converted_instrument_bytes(): array
    {
        $ntp_instrument_bytes = [];

        foreach ($this->data as $byte) {
            /* shift waveform, the GS has 128 for silence, the Amiga has 0 */
            $ensoniq_byte = 255 & ($byte + 128);

            /* make sure we do not have a 0 in the wave, which would stop the ensoniq from playing */
            if ($ensoniq_byte === 0) {
                $ensoniq_byte = 1;
            }

            $ntp_instrument_bytes[] = $ensoniq_byte;
        }

        return $ntp_instrument_bytes;
    }

    public function must_be_streamed(): bool
    {
        /* if this is set up to be streamed, return true */
        if ($this->must_stream) {
            return true;
        }

        /* if it fits exactly, docram is possible */
        if ($this->length === 32768) {
            return true;
        }

        /* if it is longer than 32k, it must be streamed */
        /* if it is < 32k it would fit into docram, but we would need the stopper bytes */
        return ($this->length > 32768 - self::STOPPER_BYTES_COUNT);
    }

    public function force_stream()
    {
        $this->must_stream = true;
    }
}

class ntpinstrument
{
    /**
     * type is a 3 bits field: stt
     * s = stream 0=no 1=yes
     * tt = type single/looped/loopheader/loop
     * 
     * if s=1 then tt must be 0 or 1
     */
    public const TYPE_SINGLE = 0;
    public const TYPE_LOOPED = 1;
    public const TYPE_LOOPHEADER = 2;
    public const TYPE_LOOP = 3;
    public const TYPE_STREAM_SINGLE = 8;
    public const TYPE_STREAM_LOOPED = 9;

    public const OPTIMUM_LENGTHS = [256, 512, 1024, 2048, 4096, 8192, 16384, 32768];

    public const STOPPER_BYTES_COUNT = 8;

    private const OPTIMUM_SMALL_LOOP_LENGTH = 512; /* a comfortable length if the instrument is a small loop */

    private const INITIAL_STREAMBUFFER_LENGTH = 512;
    private const STREAMBUFFER_LENGTH = 256;
    private const MINIMAL_LENGTH = 256;

    public $type = 0;
    public $number = 0;
    public $number_original = 0; /* this maps to the original instrument number for reference */
    public $name = '';
    public $original_length = 0; /* this is the raw length without stopper bytes! */
    public $volume = 0;
    public $finetune = 0;
    public $repeat_pointer = 0;
    public $repeat_length = 0;
    public $data = [];
    public $length = 0;
    public $page_size = 0;
    private $has_stopper_bytes = false;

    private $data_original_converted = []; /* ensoniq compatible bytes from original instrument */

    public function __construct(modinstrument $modinstrument, int $type, int $instrument_number, bool $rotate_loops)
    {
        $this->type = $type;
        $this->number = $instrument_number;

        $this->number_original = $modinstrument->number;
        $this->name = $modinstrument->name;
        $this->volume = $modinstrument->volume;
        $this->finetune = $modinstrument->finetune;
        $this->repeat_pointer = $modinstrument->repeat_pointer;
        $this->repeat_length = $modinstrument->repeat_length;
        $this->data_original_converted = $modinstrument->get_converted_instrument_bytes();
        $this->original_length = count($this->data_original_converted);

        $this->data = $this->get_instrument_data($rotate_loops);
        $this->length = count($this->data);
        $this->page_size = ceil($this->length / 256); /* how many pages does this instrument need? */
    }

    private function is_optimal_size(array $data): bool
    {
        return in_array(count($data), self::OPTIMUM_LENGTHS);
    }

    private function optimize_loop($data): array
    {
        $length = count($data);
        $has_optimal_size = in_array($length, self::OPTIMUM_LENGTHS);

        /* if the loop has not an ideal length, and is very short, copy the loop to improve sound quality */
        if (!$has_optimal_size && $length < self::OPTIMUM_SMALL_LOOP_LENGTH) {
            $number_of_stopper_bytes = self::STOPPER_BYTES_COUNT;

            /* edge case: if the loop fits exactly n times into the optimal length */
            $factor = self::OPTIMUM_SMALL_LOOP_LENGTH / $length;
            if ($factor == floor($factor)) {
                $has_optimal_size = true;
                $number_of_stopper_bytes = 0;
            }

            $loop_data = $data;
            while (count($data) + count($loop_data) + $number_of_stopper_bytes <= self::OPTIMUM_SMALL_LOOP_LENGTH) {
                $data = array_merge($data, $loop_data);
            }
        }

        return $data;
    }

    private function find_closest_zero_crossing(array $data): int
    {
        $min_ptr = 0;
        $min_diff = 1000;
        foreach ($data as $ptr => $byte) {
            if ($min_diff > abs($byte - 128)) {
                $min_diff = abs($byte - 128);
                $min_ptr = $ptr;
            }
        }
        return $min_ptr;
    }

    private function split_rotate_loop($rotate_loops): array
    {
        $data_head = array_slice($this->data_original_converted, 0, $this->repeat_pointer);
        $data_loop = array_slice($this->data_original_converted, $this->repeat_pointer, $this->repeat_length);

        $position = $this->find_closest_zero_crossing($data_loop);

        if(!$rotate_loops || $position === 0) {
            $data_head_new = $data_head;
            $data_loop_new = $data_loop;
        } else {
            $data_head_new = array_merge(
                $data_head,                           /* keep head */
                array_slice($data_loop, 0, $position) /* add beginning of loop */
            );
            $data_loop_new = array_merge(
                array_slice($data_loop, $position),   /* start loop from position to the end */
                array_slice($data_loop, 0, $position) /* add start of loop from beginning to position */
            );
        }

        /* when head + loop <= 256, append loop if possible */
        $head_rest = self::MINIMAL_LENGTH - count($data_head_new);
        if($head_rest > 0) {
            $length_loop = count($data_loop_new);
            $stopper_bytes = self::STOPPER_BYTES_COUNT;
            if(fmod($head_rest, $length_loop) == 0) {
                $stopper_bytes = 0; /* optimum fit */
            }
            /*while(count($data_head_new) + $length_loop + $stopper_bytes < self::MINIMAL_LENGTH) {
                $data_head_new = array_merge($data_head_new, $data_loop_new);
            }*/
        }

        return [$data_head_new, $data_loop_new];
    }

    /**
     * depending on the type of NTP instrument:
     * 
     * - get data from original wave
     * - optimize loops if they are very short
     * - rotate loops if we're ordered to do
     * - add stopper bytes
     */
    private function get_instrument_data(bool $rotate_loops): array
    {
        switch ($this->type) {
            case self::TYPE_SINGLE:
                $data = $this->data_original_converted;
                $add_stopper_bytes = !$this->is_optimal_size($data);
                break;

            case self::TYPE_LOOPED:
                $data = array_slice($this->data_original_converted, $this->repeat_pointer, $this->repeat_length);
                $data = $this->optimize_loop($data);
                $add_stopper_bytes = !$this->is_optimal_size($data);
                break;

            case self::TYPE_LOOPHEADER:
                [$head, $loop] = $this->split_rotate_loop($rotate_loops);
                $data = $head;
                $add_stopper_bytes = !$this->is_optimal_size($data);
                break;

            case self::TYPE_LOOP:
                [$head, $loop] = $this->split_rotate_loop($rotate_loops);
                $data = $loop;
                $data = $this->optimize_loop($data);
                $add_stopper_bytes = !$this->is_optimal_size($data);
                break;

            case self::TYPE_STREAM_SINGLE;
                $data = $this->data_original_converted;
                $add_stopper_bytes = true;
                break;

            case self::TYPE_STREAM_LOOPED;
                $head = array_slice($this->data_original_converted, 0, $this->repeat_pointer);
                $loop = array_slice($this->data_original_converted, $this->repeat_pointer, $this->repeat_length);
                $loops = $loop;
                while(count($loops) < self::INITIAL_STREAMBUFFER_LENGTH) {
                    /* for small streamed instruments, make sure the loops are larger than 512 bytes */
                    $loops = array_merge($loops, $loop);
                }
                /* always append 256 bytes of the loop so that we can always copy 256 bytes when streaming */
                $loops = array_merge($loops, array_slice($loops, 0, self::STREAMBUFFER_LENGTH));
                $data = array_merge($head, $loops);
                $add_stopper_bytes = false;
                break;
        }

        $this->has_stopper_bytes = $add_stopper_bytes;
        if ($add_stopper_bytes) {
            for ($i = 0; $i < self::STOPPER_BYTES_COUNT; $i++) {
                $data[] = 0;
            }
        }

        return $data;
    }

    public function is_first_splitted_instrument(): bool
    {
        return ($this->type === self::TYPE_LOOPHEADER);
    }

    public function is_streamed(): bool
    {
        return in_array($this->type, [self::TYPE_STREAM_SINGLE, self::TYPE_STREAM_LOOPED]);
    }

    public function get_type_for_ntp(): int
    {
        $optimal_size = $this->is_optimal_size($this->data);
        return $this->type + (($optimal_size && !$this->has_stopper_bytes) ? 1 : 0) * 4;
    }

    public function get_osc_count(): int
    {
        $optimal_size = $this->is_optimal_size($this->data);

        if($this->type === self::TYPE_SINGLE
            || ($this->type === self::TYPE_LOOPED && $optimal_size && !$this->has_stopper_bytes)
        ) {
            return 1;
        }

        return 2;
    }
}

class instrumentcontainer
{
    private $doc_pages = [];                /* a list of free/blocked 256 pages of the doc ram */
    private $doc_ram = [];                  /* the 64k doc ram */
    private $stuff_for_docram = [];

    private $ntpinstruments = [];           /* the list of ntp instruments to work with */
    private $number_of_tracks = 4;          /* if we need to reserve streambuffers, we need to know for how many tracks */
    private $track_mod_instruments = [];    /* if we need to reserve streambuffers, find out how many tracks need to have one */

    private $track_pages = [];

    private $instrument_errors = [];        /* a list of errors that occurred when trying to place instruments in doc ram */
    private $instrument_doc_pages = [];
    private $instrument_doc_sizes = [];

    public const DOC_PAGE_BOUNDARIES = [    /* instruments in the DOC need to be aligned to this number of pages */
        1,
        2,
        4,
        8,
        16,
        32,
        64,
        128,
    ];

    public const DOC_REGISTER_SIZES = [     /* the doc wavetable size register is set according to the wave size */
        0b000000,
        0b001001,
        0b010010,
        0b011011,
        0b100100,
        0b101101,
        0b110110,
        0b111111,
    ];

    private const STREAMBUFFER_LENGTH = 512;        /* the length of a stream buffer */

    public const TYPE_INSTRUMENT = 1;
    public const TYPE_STREAMBUFFER = 2;
    public const TYPE_INTERRRUPT_INSTRUMENT = 3;

    public function __construct(array $ntpinstruments, int $number_of_tracks, array $track_mod_instruments)
    {
        $this->ntpinstruments = $ntpinstruments;
        $this->number_of_tracks = $number_of_tracks;
        $this->track_mod_instruments = $track_mod_instruments;
    }

    /**
     * Helper method for get_instrument_error()
     */
    private function get_smaller_length($length): int
    {
        foreach (array_reverse(self::DOC_PAGE_BOUNDARIES) as $page_boundary) {
            $byte_length = $page_boundary * 256;
            if ($byte_length < $length) {
                return $byte_length;
            }
        }

        return $length;
    }

    /**
     * Helper method to generate a hopefully helpful message to the end user.
     */
    private function get_instrument_error(ntpinstrument $ntpinstrument, string $error): string
    {
        /* from the GS instrument number, find the original instrument number */
        $instrument_number_original = $ntpinstrument->number_original;

        /* Give a hint about making the instrument smaller */
        $sizehint = '';
        switch ($ntpinstrument->type) {
            default:
            case ntpinstrument::TYPE_SINGLE:
            case ntpinstrument::TYPE_STREAM_SINGLE:
                $length = $ntpinstrument->original_length;
                $sizehint = 'consider reducing the size from %s to %s bytes';
                break;
            case ntpinstrument::TYPE_LOOPHEADER:
                $length = $ntpinstrument->length;
                $sizehint = 'consider reducing the loop header from %s to %s bytes or removing it';
                break;
            case ntpinstrument::TYPE_LOOPED:
            case ntpinstrument::TYPE_LOOP:
            case ntpinstrument::TYPE_STREAM_LOOPED:
                $length = $ntpinstrument->length;
                $sizehint = 'consider reducing the loop from %s to %s bytes';
                break;
        }

        $smaller_length = $this->get_smaller_length($length);
        if ($smaller_length < $length) {
            $sizehint = ' - ' . sprintf($sizehint, $length, $smaller_length);
        }

        return sprintf('Instrument %s: %s%s.', str_pad($instrument_number_original, 3, '0', STR_PAD_LEFT), $error, $sizehint);
    }

    /**
     * For a given page size, reserve memory in doc ram.
     * 
     * If successful, return the page where the newly reserved memory begins.
     * If unsuccessful, return a negative value.
     */
    private function reserve_space_in_docram(int $page_size, int $type): int
    {
        /* init */
        $align_boundary = -1;
        $page_start = -1;

        /* Depending on the size of an instrument, find the matching alignment.
           For example, an instrument with length of 423 takes up 2 pages (=512 bytes) and therefore needs to be aligned on 512 bytes.
        */
        foreach (self::DOC_PAGE_BOUNDARIES as $i => $page_limit) {
            if ($page_size <= $page_limit) {
                $align_boundary = $page_limit;
                break;
            }
        }

        if ($align_boundary === -1) {
            /* If we don't find a matching alignment, it means that the instrument is larger than 32k and the DOC can't play that */
            return -2;
        }

        /* find a free slot in the DOC ram */
        $step_size = $align_boundary;
        for ($i = 0; $i < 256; $i += $step_size) {
            if ($this->doc_pages[$i] === 0) {
                /* we found a free slot */
                $page_start = $i;
                break;
            }
        }

        if ($page_start !== -1) {
            /* free slot found; mark DOC ram as occupied */
            for ($i = $page_start; $i < $page_start + $page_size; $i++) {
                $this->doc_pages[$i] = $type;
            }
        }

        return $page_start;
    }

    /**
     * Get the matching doc size register for a given page size
     */
    private function find_doc_register_size($page_size): int
    {
        if ($page_size === 0) {
            return 0;
        }

        /* Find DOC Register size */
        $doc_register_size = -1;
        foreach (self::DOC_PAGE_BOUNDARIES as $i => $page_limit) {
            if ($page_size <= $page_limit) {
                $doc_register_size = self::DOC_REGISTER_SIZES[$i];
                break;
            }
        }

        return $doc_register_size;
    }

    /**
     * Copy bytes into doc ram
     */
    private function copy_bytes_into_docram(int $page_start, array $bytes)
    {
        $start = $page_start * 256;

        $i = 0;
        foreach ($bytes as $byte) {
            $this->doc_ram[$start + $i] = $byte;
            $i++;
        }
    }

    /**
     * Try to put the interrupt instrument into doc ram
     * 
     * if this is not possible, record error(s)
     */
    private function put_interrupt_instrument_into_docram(int $page_start)
    {
        if ($page_start >= 0) {
            $this->copy_bytes_into_docram($page_start, array_fill(0, 256, 128));
        } else {
            $this->instrument_errors[] = 'Unable to place interrupt dummy instrument in doc ram.';
        }
    }

    /**
     * Try to put an instrument into doc ram
     * 
     * if this is not possible, record error(s)
     */
    private function put_instrument_in_docram(int $page_start, ntpinstrument $ntpinstrument)
    {
        if ($page_start === -2) {
            /* If we don't find a matching alignment, it means that the instrument is larger than 32k and the DOC can't play that */
            $error = 'Too large (must be < 32k)';
            $this->instrument_errors[] = $this->get_instrument_error($ntpinstrument, $error);
        } elseif ($page_start === -1) {
            /* if we don't find a free slot in the DOC ram */
            $error = 'Too large';
            $this->instrument_errors[] = $this->get_instrument_error($ntpinstrument, $error);
        } else {
            $this->copy_bytes_into_docram($page_start, $ntpinstrument->data);
            $this->instrument_doc_pages[$ntpinstrument->number] = $page_start;
            $this->instrument_doc_sizes[$ntpinstrument->number] = $this->find_doc_register_size($ntpinstrument->page_size);
        }
    }

    /**
     * Try to put a streambuffer into doc ram
     * 
     * if this is not possible, record error(s)
     */
    private function put_streambuffer_into_docram(int $page_start, int $track_number)
    {
        if ($page_start >= 0) {
            $this->track_pages[$track_number] = $page_start;
        } else {
            $this->instrument_errors[] = 'Unable to allocate doc ram memory for stream buffer.';
        }
    }

    /**
     * Try to place all the stuff we have into doc ram.
     */
    private function put_stuff_in_docram()
    {
        /* for each page (a 256 byte block), we mark if it is used or not; init with 0 */
        $this->doc_pages = array_fill(0, 256, 0);

        /* clear doc ram */
        $this->doc_ram = array_fill(0, 65536, 0);

        /* clear values of where instruments are placed and their doc size */
        $this->doc_ptrs = [];

        /* init pages for which track uses which stream buffer */
        $this->track_pages = array_fill(0, $this->number_of_tracks, 0);

        foreach ($this->stuff_for_docram as $index => $stuff) {
            $size = $stuff['size'];
            $page_size = ceil($size / 256);

            /* get page start, negative values mean an error occurred */
            $page_start = $this->reserve_space_in_docram($page_size, $stuff['type']);

            /* record page where the instrument is placed at */
            $this->stuff_for_docram[$index]['page_start'] = $page_start;

            switch ($stuff['type']) {
                case self::TYPE_INSTRUMENT:
                    $this->put_instrument_in_docram($page_start, $stuff['instrument']);
                    break;
                case self::TYPE_STREAMBUFFER:
                    $this->put_streambuffer_into_docram($page_start, $stuff['track_number']);
                    break;
                case self::TYPE_INTERRRUPT_INSTRUMENT:
                    $this->put_interrupt_instrument_into_docram($page_start);
                    break;
            }
        }
    }

    /**
     * Rare edge case:
     *
     * Make sure that the first 256 of the DOC ram are filled with non-zeros.
     * Otherwise, the interrupt timer cannot work.
     * 
     * Since the largest non-streamed instrument will be placed to this address, this is
     * extremely unlikely to happen.
     * 
     * But if it does, we add a dummy 256 byte silent instrument.
     */
    private function must_reserve_for_interrupt_instrument(): bool
    {
        $max_instrument_length = 0;
        foreach ($this->ntpinstruments as $ntpinstrument) {
            /** @var ntpinstrument $ntpinstrument */
            if(!$ntpinstrument->is_streamed()) {
                $max_instrument_length = max($max_instrument_length, $ntpinstrument->length);
            }
        }

        if ($max_instrument_length < 256) {
            /* even the largest instrument is too small, we have to reserve space for the interrupt */
            return true;
        }

        return false;
    }

    /**
     * Determine what has to be placed in doc ram. There can be:
     * 
     * - a dummy 256 byte silence for the interrupt timer
     * - instruments that are not streamed
     * - buffers for streaming samples (only if there are samples to stream)
     */
    private function get_stuff_for_docram()
    {
        $stuff_for_docram = [];

        if ($this->must_reserve_for_interrupt_instrument()) {
            $stuff_for_docram[] = [
                'type' => self::TYPE_INTERRRUPT_INSTRUMENT,
                'size' => 256,
                'pages' => 1,
            ];
        }

        $streaming_tracks = [];
        foreach ($this->ntpinstruments as $ntpinstrument) {
            /** @var ntpinstrument $ntpinstrument */
            if ($ntpinstrument->is_streamed()) {
                /* flag all tracks with this streaming instrument */
                foreach ($this->track_mod_instruments as $track => $mod_numbers) {
                    if (in_array($ntpinstrument->number_original, $mod_numbers)) {
                        $streaming_tracks[$track] = 1;
                    }
                }
            } else {
                /* else add non-streamed instrument */
                $stuff_for_docram[] = [
                    'type' => self::TYPE_INSTRUMENT,
                    'size' => $ntpinstrument->length,
                    'instrument' => $ntpinstrument,
                    'pages' => $ntpinstrument->page_size,
                ];
            }
        }

        foreach ($streaming_tracks as $track_number => $flag) {
            /* add stream buffers only for tracks that play a streamed sample */
                $stuff_for_docram[] = [
                    'type' => self::TYPE_STREAMBUFFER,
                    'size' => self::STREAMBUFFER_LENGTH,
                    'track_number' => $track_number,
                    'pages' => self::STREAMBUFFER_LENGTH / 256,
                ];
            }

        $this->stuff_for_docram = $stuff_for_docram;
    }

    /**
     * Helper method to sort the stuff to put into doc ram
     * TYPE_INTERRRUPT_INSTRUMENT has highest priority though
     */
    private function sort_stuff_for_docram_by_size(array $a, array $b): int
    {
        if($a['type'] === self::TYPE_INTERRRUPT_INSTRUMENT) {
            return -1;
        }

        if($b['type'] === self::TYPE_INTERRRUPT_INSTRUMENT) {
            return 1;
        }

        if ($a['pages'] === $b['pages']) {
            return 0;
        }

        return $a['pages'] > $b['pages'] ? -1 : 1;
    }

    /**
     * This is the main method to manage instruments and doc ram
     */
    public function arrange_instruments(): bool
    {
        /* determine what we should try to put into the doc ram */
        $this->get_stuff_for_docram();

        /* sort by size, largest first */
        usort($this->stuff_for_docram, [$this, 'sort_stuff_for_docram_by_size']);

        /* try to put stuff in docram */
        $this->put_stuff_in_docram();

        /* report if there was an error */
        return (count($this->instrument_errors) === 0);
    }

    /**
     * Get instrument errors
     */
    public function get_instrument_errors(): array
    {
        return $this->instrument_errors;
    }

    public function get_stream_track_pages(): array
    {
        return $this->track_pages;
    }

    public function get_instrument_doc_ptrs(): array
    {
        foreach ($this->ntpinstruments as $ntpinstrument) {
            if (!isset($this->instrument_doc_pages[$ntpinstrument->number])) {
                $this->instrument_doc_pages[$ntpinstrument->number] = 0;
            }
        }

        return $this->instrument_doc_pages;
    }

    public function get_instrument_doc_sizes(): array
    {
        foreach ($this->ntpinstruments as $ntpinstrument) {
            if (!isset($this->instrument_doc_sizes[$ntpinstrument->number])) {
                $this->instrument_doc_sizes[$ntpinstrument->number] = 0;
            }
        }

        return $this->instrument_doc_sizes;
    }

    public function get_doc_ram_usage()
    {
        return $this->doc_pages;
    }
}

class ntpconverter
{
    public const CONFIG_STREAM_FORBIDDEN = 0;
    public const CONFIG_STREAM_ALLOW = 1;
    public const CONFIG_STREAM_ENFORCE = 2;

    public static $FATAL_ERROR = 1000;

    private $path_source_mod;          /* path to the MOD file that should be converted */
    private $path_output_ntp;          /* path to the resulting ninjatracker+ (.ntp) file */

    private $stream_config;            /* true if we allow streamed instruments */

    private $mod_data;                 /* the binary file data of the MOD */
    private $mod_name;                 /* the name string of the MOD */
    private $max_number_instruments;   /* how many instruments this MOD has at max */
    private $no_of_tracks;             /* how many tracks this MOD uses */
    private $max_pat_cnt;              /* number of patterns (blocks) that the MOD has */
    private $ptr_instnumb;             /* pointer to where instruments are in a MOD */
    private $mod_headlen;              /* length of the MOD header */
    private $ptr_songp_start;          /* pointer to where the Songinfo is, which block should be played after another */
    private $ptr_songp_end;            /* pointer to where the Songinfo ends */
    private $length_of_one_pattern;    /* how many bytes form a pattern in this MOD */
    private $length_of_all_patterns;   /* how many bytes form all patterns in this MOD, combined */
    private $pattern_order;            /* array of pattern numbers that defines the order in which to play the patterns */

    private $needs_stream;             /* array of instrument numbers that have to be streamed because of the way they are used */
    private $track_mod_instruments;    /* array of mod instrument numbers for each track */

    private $mod_instruments;          /* array of original instruments */
    private $ntp_instruments;          /* array of instruments for the GS */

    private $doc_ptrs;                 /* array of pointers to where the GS instrument are placed in the DOC */
    private $doc_sizes;                /* array of size values that can be used by the DOC to play the instrument */

    private $track_stream_pages;       /* array of pointers to where, for each track, a stream buffer is placed in the DOC */
    private $track_oscillator_byte;    /* array of bytes %x00yyyyy with x=oscillators to use (0=1, 1=2), yyyyy: oscillator to use */

    private $ntp_notes;                /* array of ntp pattern data bytes */

    private $ntp_instrument_positions; /* array of pointers in file where each instruments is placed */

    private $instrument_errors;        /* array of errors when converting instruments */
    private $pattern_errors;           /* array of errors when converting patterns */
    private $track_errors;             /* array of errors when converting tracks */

    private instrumentcontainer $instrumentcontainer;

    private static $ptr_first_instrument = 20;       /* pointer to where the MOD has its first instrument */
    private static $length_instrument_info = 30;     /* length of an instrument information block in the MOD */
    private static $offset_instrument_length = 22;   /* offset in instrument information block to the length of the instrument */
    private static $offset_instrument_finetune = 24; /* offset in instrument information block to the finetune value */
    private static $offset_instrument_volume = 25;   /* offset in instrument information block to the volume */
    private static $offset_instrument_rep_ptr = 26;  /* offset in instrument information block to the repeat pointer within the instrument */
    private static $offset_instrument_rep_len = 28;  /* offset in instrument information block to the length of the repeat */
    private static $ptr_mod_type = 1080;             /* pointer to where the MOD identifier code is */

    private static $ptr_by_type = [
        'old' => [ /* old trackers have up to 15 instruments */
            'max_number_instruments' => 15,
            'ptr_instnumb' => 950 - 480,
            'mod_headlen' => 1080 - 480,
            'ptr_songp_start' => 952 - 480,
            'ptr_songp_end' => 1080 - 480,
        ],
        'normal' => [ /* normal trackers have up to 31 instruments */
            'max_number_instruments' => 31,
            'ptr_instnumb' => 950,
            'mod_headlen' => 1084,
            'ptr_songp_start' => 952,
            'ptr_songp_end' => 1080,
        ],
    ];
    private static $length_of_note = 4;              /* 4 bytes represent a note in a mod */
    private static $number_of_lines_in_pattern = 64; /* there are always 64 lines in a pattern */
    private static $max_ntp_instruments = 255;       /* there may be only a limited number of instruments in an ntp */

    private static $mod_types = [ /* mapping for ID in MOD => number of tracks */
        'M.K.' => 4,
        '1CHN' => 1,
        '2CHN' => 2,
        '3CHN' => 3,
        '4CHN' => 4,
        '5CHN' => 5,
        '6CHN' => 6,
        '7CHN' => 7,
        '8CHN' => 8,
        '9CHN' => 9,
        '01CH' => 1,
        '02CH' => 2,
        '03CH' => 3,
        '04CH' => 4,
        '05CH' => 5,
        '06CH' => 6,
        '07CH' => 7,
        '08CH' => 8,
        '09CH' => 9,
        '10CH' => 10,
        '11CH' => 11,
        '12CH' => 12,
        '13CH' => 13,
        '14CH' => 14,
        '15CH' => 15,
        '16CH' => 16,
        '17CH' => 17,
        '18CH' => 18,
        '19CH' => 19,
        '20CH' => 20,
        '21CH' => 21,
        '22CH' => 22,
        '23CH' => 23,
        '24CH' => 24,
        '25CH' => 25,
        '26CH' => 26,
        '27CH' => 27,
        '28CH' => 28,
        '29CH' => 29,
        '30CH' => 30,
        '31CH' => 31,
        '01CN' => 1,
        '02CN' => 2,
        '03CN' => 3,
        '04CN' => 4,
        '05CN' => 5,
        '06CN' => 6,
        '07CN' => 7,
        '08CN' => 8,
        '09CN' => 9,
        '10CN' => 10,
        '11CN' => 11,
        '12CN' => 12,
        '13CN' => 13,
        '14CN' => 14,
        '15CN' => 15,
        '16CN' => 16,
        '17CN' => 17,
        '18CN' => 18,
        '19CN' => 19,
        '20CN' => 20,
        '21CN' => 21,
        '22CN' => 22,
        '23CN' => 23,
        '24CN' => 24,
        '25CN' => 25,
        '26CN' => 26,
        '27CN' => 27,
        '28CN' => 28,
        '29CN' => 29,
        '30CN' => 30,
        '31CN' => 31,
        'FLT4' => 4,
        'FLT8' => 8,
    ];

    private static $mod_notes = [ /* music notes that are possible in the MOD */
        1712, 1616, 1524, 1440, 1356, 1280, 1208, 1140, 1076, 1016, 960, 906, /* c2-b2 */
        856,   808,  762,  720,  678,  640,  604,  570,  538,  508, 480, 453, /* c3-b3 */
        428,   404,  381,  360,  339,  320,  302,  285,  269,  254, 240, 226, /* c4-b4 */
        214,   202,  190,  180,  170,  160,  151,  143,  135,  127, 120, 113, /* c5-b5 */
        107,   101,   95,   90,   85,   80,   75,   71,   67,   63,  60,  56, /* c6-b6 */
        53,     50,   47,   45,   42,   40,   37,   35,   33,   31,  30,  28, /* c7-b7 */
    ];

    private static $doc_stereo_defaults = [
        1, 0,
        0, 1,
        1, 0,
        0, 1,
        1, 0,
        0, 1,
        1, 0,
        0, 1,
        1, 0,
        0, 1,
        1, 0,
        0, 1,
        1, 0,
        0, 1,
        1, 0,
        0, 1,
    ];

    /* ---- helper methods ---- */

    private function get_mod_file_byte($position)
    {
        return ord(substr($this->mod_data, $position, 1));
    }

    private function get_amiga_word($position)
    {
        return 256 * $this->get_mod_file_byte($position) + $this->get_mod_file_byte($position + 1);
    }

    private function length_to_gs_str($word)
    {
        $vhi_byte = ($word & 0xff0000) >> 16;
        $hi_byte = ($word & 0xff00) >> 8;
        $lo_byte = $word & 0x00ff;

        return chr($lo_byte) . chr($hi_byte) . chr($vhi_byte);
    }

    private function get_module_string($ptr, $max_length)
    {
        $string = '';
        for ($i = 0; $i < $max_length; $i++) {
            $byte = $this->get_mod_file_byte($ptr + $i);
            if ($byte === 0) {
                break;
            }

            $string .= chr($byte);
        }
        return $string;
    }

    /* ---- main program code ---- */

    public function __construct($path_source_mod, $path_output_ntp, $stream_config = self::CONFIG_STREAM_ALLOW)
    {
        $this->path_source_mod = $path_source_mod;
        $this->path_output_ntp = $path_output_ntp;

        $this->instrument_errors = [];
        $this->pattern_errors = [];
        $this->track_errors = [];

        $this->stream_config = $stream_config;
    }

    private function detect_mod_and_number_of_tracks()
    {
        /* get module name */
        $this->mod_name = $this->get_module_string(0, 20);

        /* find out if old or normal and how many tracks there are */
        $str_type = substr($this->mod_data, self::$ptr_mod_type, 4);
        if (isset(self::$mod_types[$str_type])) {
            $old_or_normal = 'normal';
            $this->no_of_tracks = self::$mod_types[$str_type];
        } else {
            $old_or_normal = 'old';
            $this->no_of_tracks = 4;
        }

        /* set values, depending on old or normal */
        $this->max_number_instruments = self::$ptr_by_type[$old_or_normal]['max_number_instruments'];
        $this->ptr_instnumb = self::$ptr_by_type[$old_or_normal]['ptr_instnumb'];
        $this->mod_headlen = self::$ptr_by_type[$old_or_normal]['mod_headlen'];
        $this->ptr_songp_start = self::$ptr_by_type[$old_or_normal]['ptr_songp_start'];
        $this->ptr_songp_end = self::$ptr_by_type[$old_or_normal]['ptr_songp_end'];
    }

    private function detect_number_of_patterns()
    {
        $max_pat_cnt = 0;
        for ($pos = $this->ptr_songp_start; $pos < $this->ptr_songp_end; $pos++) {
            $byte = $this->get_mod_file_byte($pos);
            $max_pat_cnt = max($byte, $max_pat_cnt);
        }
        $max_pat_cnt++; /* add 1 because pattern 0 is also a pattern */

        $this->max_pat_cnt = $max_pat_cnt;
    }

    private function calc_pattern_length()
    {
        $this->length_of_one_pattern = $this->no_of_tracks * self::$length_of_note * self::$number_of_lines_in_pattern;

        $this->length_of_all_patterns = $this->length_of_one_pattern * $this->max_pat_cnt;
    }

    private function get_pattern_order()
    {
        $pattern_order_length = $this->get_mod_file_byte($this->ptr_instnumb); /* length of pattern order */
        $this->pattern_order = [];

        for ($i = 0; $i < $pattern_order_length; $i++) {
            $ptr_mod = $this->ptr_instnumb + 2 + $i;
            $byte = $this->get_mod_file_byte($ptr_mod);

            $this->pattern_order[] = $byte;
        }
    }

    private function extract_instrument($instrument_number, $total_instrument_length)
    {
        /* instrument numbers run from 1 to 15 or 1 to 31 */
        $ptr = self::$ptr_first_instrument + ($instrument_number - 1) * self::$length_instrument_info;

        $instrument_name = $this->get_module_string($ptr, 22);
        $instrument_length = $this->get_amiga_word($ptr + self::$offset_instrument_length) * 2; /* given in words, transform to bytes */
        $instrument_volume = $this->get_mod_file_byte($ptr + self::$offset_instrument_volume);
        $instrument_finetune = (15 & $this->get_mod_file_byte($ptr + self::$offset_instrument_finetune)); /* lower nibble only */
        $instrument_rep_ptr = $this->get_amiga_word($ptr + self::$offset_instrument_rep_ptr) * 2; /* given in words, transform to bytes */
        $instrument_rep_len = $this->get_amiga_word($ptr + self::$offset_instrument_rep_len) * 2; /* given in words, transform to bytes */

        if ($instrument_length === 0) {
            /* No instrument, skip */
            return false;
        }

        /* calculcate the pointer in the MOD where the instrument begins */
        $ptr = $this->mod_headlen + $this->length_of_all_patterns + $total_instrument_length;

        /* get instrument data */
        $instrument_data = [];
        for ($i = 0; $i < $instrument_length; $i++) {
            $instrument_data[] = $this->get_mod_file_byte($ptr + $i);
        }

        /* build instrument */
        $instrument = new modinstrument();
        $instrument->number = $instrument_number;
        $instrument->name = $instrument_name;
        $instrument->length = $instrument_length;
        $instrument->finetune = $instrument_finetune;
        $instrument->volume = $instrument_volume;
        $instrument->repeat_pointer = $instrument_rep_ptr;
        $instrument->repeat_length = $instrument_rep_len;
        $instrument->data = $instrument_data;

        /*
         * if the effect requires a streamed instrument, set instrument to stream
         * but when streaming is not allowed, ignore this
         */
        if (
            $this->stream_config === self::CONFIG_STREAM_ENFORCE
            ||
            (
                $this->stream_config === self::CONFIG_STREAM_ALLOW
                &&
                in_array($instrument_number, $this->needs_stream)
            )
        ) {
            $instrument->force_stream();
        }

        return $instrument;
    }

    private function extract_instruments()
    {
        $this->mod_instruments = [];

        $total_instrument_length = 0;
        for ($i = 1; $i <= $this->max_number_instruments; $i++) {
            $instrument = $this->extract_instrument($i, $total_instrument_length);
            if ($instrument !== false) {
                /** @var modinstrument $instrument */
                $this->mod_instruments[$i] = $instrument;
                $total_instrument_length += $instrument->length;

                /*
                echo "num " . str_pad($instrument->number, 3, '0', STR_PAD_LEFT);
                echo ", type=" . $instrument->get_instrument_type();
                echo ", streamed=" . ($instrument->must_be_streamed() ? 'yes' : 'no ');
                echo ", len=" . str_pad($instrument->length, 5, '0', STR_PAD_LEFT);
                echo ", vol=" . str_pad($instrument->volume, 2, '0', STR_PAD_LEFT);
                echo ", finetune=" . str_pad($instrument->finetune, 2, ' ', STR_PAD_LEFT);
                echo "\n";
                */
            }
        }
    }

    private function convert_instruments(bool $rotate_loops = false)
    {
        $this->ntp_instruments = [];

        $ntp_instrument_number = 1;
        foreach ($this->mod_instruments as $modinstrument) {

            /** @var modinstrument $modinstrument */

            if (!$modinstrument->must_be_streamed()) {
                switch ($modinstrument->get_instrument_type()) {
                    case modinstrument::TYPE_SINGLE:
                        $this->ntp_instruments[] = new ntpinstrument($modinstrument, ntpinstrument::TYPE_SINGLE, $ntp_instrument_number, $rotate_loops);
                        break;

                    case modinstrument::TYPE_LOOPED:
                        $this->ntp_instruments[] = new ntpinstrument($modinstrument, ntpinstrument::TYPE_LOOPED, $ntp_instrument_number, $rotate_loops);
                        break;

                    case modinstrument::TYPE_LOOPWITHHEADER:
                        $this->ntp_instruments[] = new ntpinstrument($modinstrument, ntpinstrument::TYPE_LOOPHEADER, $ntp_instrument_number, $rotate_loops);
                        $ntp_instrument_number++;
                        $this->ntp_instruments[] = new ntpinstrument($modinstrument, ntpinstrument::TYPE_LOOP, $ntp_instrument_number, $rotate_loops);
                        break;
                }
            } else {
                switch ($modinstrument->get_instrument_type()) {
                    case modinstrument::TYPE_SINGLE:
                        $this->ntp_instruments[] = new ntpinstrument($modinstrument, ntpinstrument::TYPE_STREAM_SINGLE, $ntp_instrument_number, $rotate_loops);
                        break;
                    default:
                        $this->ntp_instruments[] = new ntpinstrument($modinstrument, ntpinstrument::TYPE_STREAM_LOOPED, $ntp_instrument_number, $rotate_loops);
                        break;
                }
            }

            $ntp_instrument_number++;
        }
    }

    private function check_instrument_count()
    {
        if (count($this->ntp_instruments) > self::$max_ntp_instruments) {
            $this->instrument_errors[] = 'Too many instruments:';
            $this->instrument_errors[] = sprintf('- Ninjatracker only allows up to %s instruments.', self::$max_ntp_instruments);

            $number_of_mod_instruments = count($this->mod_instruments);
            if ($number_of_mod_instruments > self::$max_ntp_instruments) {
                $this->instrument_errors[] = sprintf('- The MOD already has %s instruments.', $number_of_mod_instruments);
            } else {
                $this->instrument_errors[] = sprintf('- The MOD has %s instruments.', $number_of_mod_instruments);
            }

            $number_splitted = 0;
            foreach ($this->ntp_instruments as $ntp_instrument) {
                /** @var ntpinstrument $ntp_instrument */
                $number_splitted += $ntp_instrument->is_first_splitted_instrument() ? 1 : 0;
            }

            if ($number_splitted > 0) {
                $this->instrument_errors[] = sprintf(
                    '- Because of looped instruments, the converter creates %s additional instrument%s.',
                    $number_splitted,
                    ($number_splitted != 1) ? 's' : ''
                );
            }

            $this->instrument_errors[] = '- Consider getting rid of instruments or try removing instrument loop headers.';
        }
    }

    /**
     * Switch the largest instrument from non-streaming to stream
     */
    private function switch_instrument_to_stream(): bool
    {
        /* find largest instrument */
        $max_length = 0;
        $instrument_index = -1;
        foreach ($this->mod_instruments as $index => $mod_instrument) {
            /** @var modinstrument $mod_instrument */
            if (!$mod_instrument->must_be_streamed() && $mod_instrument->length > $max_length) {
                $max_length = $mod_instrument->length;
                $instrument_index = $index;
            }
        }

        if ($instrument_index >= 0) {
            /* if we have found an instrument to switch, do it */
            $this->mod_instruments[$instrument_index]->force_stream();
            return true;
        }

        /* no instrument available to switch, this is a fatal error */
        return false;
    }

    private function optimize_instruments_sub(bool $rotate_loops): bool
    {
        while (true) {
            $this->convert_instruments($rotate_loops);
            $this->check_instrument_count();

            if (count($this->instrument_errors) > 0) {
                return false;
            }

            $this->instrumentcontainer = new instrumentcontainer($this->ntp_instruments, $this->no_of_tracks, $this->track_mod_instruments);
            if ($this->instrumentcontainer->arrange_instruments()) {
                /* success, copy what is placed where */
                $this->track_stream_pages = $this->instrumentcontainer->get_stream_track_pages();
                $this->doc_ptrs = $this->instrumentcontainer->get_instrument_doc_ptrs();
                $this->doc_sizes = $this->instrumentcontainer->get_instrument_doc_sizes();
                return true;
            }

            /* an error occurred */
            if ($this->stream_config === self::CONFIG_STREAM_FORBIDDEN) {
                /* if streaming is not allowed, end here */
                $this->instrument_errors = array_merge($this->instrument_errors, $this->instrumentcontainer->get_instrument_errors());
                return false;
            }

            /* switch the largest not-streamed instrument to stream and try again */
            $this->switch_instrument_to_stream();
        }
    }

    private function optimize_instruments()
    {
        /* try to convert instruments using rotated loops */
        if ($this->optimize_instruments_sub(true)) {
            /* all okay with rotated loops, we're done */
            return;
        }

        $this->optimize_instruments_sub(false);
    }

    private function convert_period($period)
    {
        if ($period == 0) {
            return 0; /* no note */
        }

        $key_mod_note = array_search($period, self::$mod_notes);
        if ($key_mod_note === false) {
            throw new Exception(sprintf('Unable to convert note period value of %s.', $period));
        }

        return $key_mod_note + 1;
    }

    private function convert_instrument_number($mod_instrument_number)
    {
        if ($mod_instrument_number === 0) {
            return 0; /* no instrument */
        }

        if (isset($this->instrument_map[$mod_instrument_number])) {
            return $this->instrument_map[$mod_instrument_number];
        }

        throw new Exception(sprintf('Instrument %s not found!', $mod_instrument_number));
    }

    private function extrad_mod_note($ptr_to_note)
    {
        /* 4 bytes are one note/effect */
        $word0 = $this->get_amiga_word($ptr_to_note);
        $word1 = $this->get_amiga_word($ptr_to_note + 2);

        /* extract mod data */
        $mod_period = $word0 & 0x0fff;
        $mod_instrument_number = (($word0 & 0xf000) >> 8) + (($word1 & 0xf000) >> 12);
        $mod_effect_number = ($word1 & 0x0f00) >> 8;
        $mod_effect_value = ($word1 & 0x00ff);

        return [$mod_period, $mod_instrument_number, $mod_effect_number, $mod_effect_value];
    }

    private function convert_mod_note($ptr_to_note)
    {
        [$mod_period, $mod_instrument_number, $mod_effect_number, $mod_effect_value] = $this->extrad_mod_note($ptr_to_note);

        /* convert to gs */
        $gs_note_index = $this->convert_period($mod_period);
        $gs_instrument_number = $this->convert_instrument_number($mod_instrument_number);
        $gs_effect_number = $mod_effect_number;
        $gs_effect_value = $mod_effect_value;

        return [$gs_note_index, $gs_instrument_number, $gs_effect_number, $gs_effect_value];
    }

    private function add_note_to_ntp(array $note_data)
    {
        list($gs_note_index, $gs_instrument_number, $gs_effect_number, $gs_effect_value) = $note_data;

        /* sanity checks */
        if ($gs_note_index > 255) {
            throw new Exception('NinjaTracker+ does not allow a note value > 255!');
        }
        if ($gs_instrument_number > self::$max_ntp_instruments) {
            throw new Exception(sprintf('NinjaTracker+ does not allow more than %s instruments!', self::$max_ntp_instruments));
        }
        if ($gs_effect_number > 15) {
            throw new Exception('NinjaTracker+ does not allow effect numbers > 15!');
        }
        if ($gs_effect_value > 255) {
            throw new Exception('NinjaTracker+ does not allow effect values > 255!');
        }

        $this->ntp_notes[] = $gs_note_index;
        $this->ntp_notes[] = $gs_instrument_number;
        $this->ntp_notes[] = $gs_effect_number;
        $this->ntp_notes[] = $gs_effect_value;
    }

    private function analyze_pattern($pattern_number)
    {
        $pattern_len = $this->no_of_tracks * self::$number_of_lines_in_pattern * self::$length_of_note;
        $pattern_offset = $this->mod_headlen + $pattern_len * $pattern_number;

        for ($line = 0; $line < self::$number_of_lines_in_pattern; $line++) {
            $line_offset = $line * $this->no_of_tracks * self::$length_of_note;

            for ($track = 0; $track < $this->no_of_tracks; $track++) {
                $track_offset = $track * self::$length_of_note;

                $ptr = $pattern_offset + $line_offset + $track_offset;

                [$mod_period, $mod_instrument_number, $mod_effect_number, $mod_effect_value] = $this->extrad_mod_note($ptr);

                /* write down which track uses which instruments */
                if($mod_instrument_number > 0) {
                    if(!isset($this->track_mod_instruments[$track])) {
                        $this->track_mod_instruments[$track] = [];
                    }
                    if(!in_array($mod_instrument_number, $this->track_mod_instruments[$track])) {
                        $this->track_mod_instruments[$track][] = $mod_instrument_number;
                    }
                }

                /* check for Effect 9 */
                if ($mod_effect_number == 9 && $mod_effect_value != 0 && !in_array($mod_instrument_number, $this->needs_stream)) {
                    /* Effect 9 is used, but only works for streamed instruments! */
                    $this->needs_stream[] = $mod_instrument_number;
                }
            }
        }
    }

    private function analyze_patterns()
    {
        $this->needs_stream = [];
        $this->track_mod_instruments = [];

        for ($pattern_number = 0; $pattern_number < $this->max_pat_cnt; $pattern_number++) {
            $this->analyze_pattern($pattern_number);
        }
    }

    private function convert_pattern($pattern_number)
    {
        $pattern_len = $this->no_of_tracks * self::$number_of_lines_in_pattern * self::$length_of_note;
        $pattern_offset = $this->mod_headlen + $pattern_len * $pattern_number;

        try {
            for ($line = 0; $line < self::$number_of_lines_in_pattern; $line++) {
                $line_offset = $line * $this->no_of_tracks * self::$length_of_note;

                for ($track = 0; $track < $this->no_of_tracks; $track++) {
                    $track_offset = $track * self::$length_of_note;

                    $ptr = $pattern_offset + $line_offset + $track_offset;

                    $ntp_note_data = $this->convert_mod_note($ptr);
                    $this->add_note_to_ntp($ntp_note_data);
                }
            }
        } catch (Exception $ex) {
            $this->pattern_errors[] = sprintf(
                'Error in pattern %s, line %s, track %s: %s',
                str_pad($pattern_number, 2, '0', STR_PAD_LEFT),
                str_pad($line, 2, '0', STR_PAD_LEFT),
                str_pad($track + 1, 2, '0', STR_PAD_LEFT),
                $ex->getMessage()
            );
        }
    }

    private function convert_patterns()
    {
        /* generate instrument map */
        $this->instrument_map = [];
        foreach ($this->ntp_instruments as $ntpinstrument) {
            /** @var ntpinstrument $ntpinstrument */
            if (!isset($this->instrument_map[$ntpinstrument->number_original])) {
                /* save 1st instrument as it is the loop header */
                $this->instrument_map[$ntpinstrument->number_original] = $ntpinstrument->number;
            }
        }

        for ($pattern_number = 0; $pattern_number < $this->max_pat_cnt; $pattern_number++) {
            $this->convert_pattern($pattern_number);
        }
    }

    private function determine_osc_per_track()
    {
        /* init with 1 oscillator per track */
        $track_osc_count = [];
        for ($track = 1; $track <= $this->no_of_tracks; $track++) {
            $track_osc_count[$track] = 1;
        }

        /* find number of needed oscillators per track */
        foreach ($this->track_mod_instruments as $t => $mod_instrument_numbers) {
            $track = $t + 1;
            foreach ($mod_instrument_numbers as $mod_instrument_number) {
                foreach ($this->ntp_instruments as $ntpinstrument) {
                    /** @var ntpinstrument $ntpinstrument */
                    if ($ntpinstrument->number_original == $mod_instrument_number) {
                        $track_osc_count[$track] = max($track_osc_count[$track], $ntpinstrument->get_osc_count());
                    }
                }
            }
        }

        /* check total numer of oscillators */
        if (array_sum($track_osc_count) > 31) {
            $this->track_errors[] = $this->get_too_many_osc_error($track_osc_count);
            return;
        }

        /* assign oscillators per track */
        $track_byte = [];
        $oscs = array_fill(0, 31, 0);
        $oscs[31] = -1;
        foreach($track_osc_count as $track => $osc_needed)
        {
            for($i = 0; $i < count($oscs); $i += $osc_needed)
            {
                //find free oscillator(s)
                if($osc_needed === 1 && $oscs[$i] === 0)
                {
                    $oscs[$i] = $track; //reserve oscillator
                    $track_byte[$track] = $i;
                    break;
                }
                elseif($osc_needed === 2 && $oscs[$i] === 0 && $oscs[$i + 1] === 0)
                {
                    $oscs[$i] = $track; //reserve oscillators
                    $oscs[$i + 1] = $track;
                    $track_byte[$track] = 128 + $i;
                    break;
                }
            }
        }
        $this->track_oscillator_byte = $track_byte;
        $this->oscillator_usage = $oscs;
    }

    private function get_too_many_osc_error($track_osc_count)
    {
        foreach ($this->track_mod_instruments as $t => $mod_instrument_numbers) {
            $track = $t + 1;
            foreach ($mod_instrument_numbers as $mod_instrument_number) {
                foreach ($this->ntp_instruments as $ntpinstrument) {
                    /** @var ntpinstrument $ntpinstrument */
                    if ($ntpinstrument->number_original == $mod_instrument_number) {
                        $track_osc_count[$track] = max($track_osc_count[$track], $ntpinstrument->get_osc_count());
                    }
                }
            }
        }

        $msg = 'The tracks require too many oscillators:' . PHP_EOL;
        foreach ($track_osc_count as $track => $count) {
            $msg .= sprintf('Track %s: %s oscillator%s, instrument%s: %s' . PHP_EOL,
                str_pad($track, 2, ' ', STR_PAD_LEFT),
                $count,
                ($count === 1 ? '' : 's'),
                (count($this->track_mod_instruments[$track - 1]) === 1 ? '' : 's'),
                implode(', ', $this->track_mod_instruments[$track - 1])
            );
        }
        $msg .= sprintf('Total: %s (31 allowed).', array_sum($track_osc_count));

        $msg .= PHP_EOL . PHP_EOL;

        $msg .= $this->instrument_osc_requirements();

        return $msg;
    }

    private function instrument_osc_requirements()
    {
        $msg = 'Module instrument oscillator requirements:' . PHP_EOL;

        $instrument_osc_count = [];
        foreach ($this->ntp_instruments as $ntpinstrument) {
            $original_instrument = $ntpinstrument->number_original;
            if(!isset($instrument_osc_count[$original_instrument])) {
                $instrument_osc_count[$original_instrument] = $ntpinstrument->get_osc_count();
            }
        }

        foreach($instrument_osc_count as $instrument_number => $count) {
            $msg .= sprintf('Instrument %s: %s oscillator%s' . PHP_EOL,
                str_pad($instrument_number, 3, ' ', STR_PAD_LEFT),
                $count,
                ($count === 1 ? '' : 's')
            );
        }

        return $msg;
    }

    private function save_ntp()
    {
        /* identifier */
        $str = 'nfc!';

        /* version */
        $str .= chr(2);

        /* number of tracks */
        $str .= chr($this->no_of_tracks);

        /* number of instruments */
        $str .= chr(count($this->ntp_instruments));

        /* number of patterns */
        $str .= chr($this->max_pat_cnt);

        /* number of pattern order */
        $str .= chr(count($this->pattern_order));

        /* name */
        $str .= chr(strlen($this->mod_name));
        $str .= $this->mod_name;

        /* 3 bytes per track */
        for ($track_number = 1; $track_number <= $this->no_of_tracks; $track_number++) {
            /* 1 byte: output channel (0-7) for every track */
            $str .= chr(self::$doc_stereo_defaults[$track_number - 1]);
            /* 1 byte: page of stream buffer for each track */
            $str .= chr($this->track_stream_pages[$track_number - 1]);
            /* 1 byte: %x00yyyyy with x=oscillators to use (0=1, 1=2), yyyyy: oscillator to use */
            $str .= chr($this->track_oscillator_byte[$track_number]);
        }

        /* instrument data */
        foreach ($this->ntp_instruments as $ntp_instrument) {
            /** @var ntpinstrument $ntp_instrument */
            $str .= chr($ntp_instrument->get_type_for_ntp());
            $str .= $this->length_to_gs_str($ntp_instrument->length);
            $str .= $this->length_to_gs_str($ntp_instrument->repeat_pointer);
            $str .= $this->length_to_gs_str($ntp_instrument->repeat_length);
            $str .= chr($ntp_instrument->volume);
            $str .= chr($ntp_instrument->finetune);
            $str .= chr($this->doc_ptrs[$ntp_instrument->number]);
            $str .= chr($this->doc_sizes[$ntp_instrument->number]);
            $str .= chr(strlen($ntp_instrument->name));
            $str .= $ntp_instrument->name;
        }

        /* add pattern order */
        foreach ($this->pattern_order as $byte) {
            $str .= chr($byte);
        }

        /* add pattern data */
        foreach ($this->ntp_notes as $byte) {
            $str .= chr($byte);
        }

        /* add instrument data */
        $this->ntp_instrument_positions = [];
        foreach ($this->ntp_instruments as $ntp_instrument) {

            $this->ntp_instrument_positions[$ntp_instrument->number] = strlen($str);

            /** @var ntpinstrument $ntp_instrument */
            foreach ($ntp_instrument->data as $byte) {
                $str .= chr($byte);
            }
        }

        file_put_contents($this->path_output_ntp, $str);
    }

    private function load_mod()
    {
        $this->mod_data = file_get_contents($this->path_source_mod);

        if ($this->mod_data === false) {
            throw new Exception(sprintf('Unable to load MOD %s', $this->path_source_mod), self::$FATAL_ERROR);
        }
    }

    private function check_config()
    {
        if (!in_array($this->stream_config, [
            self::CONFIG_STREAM_FORBIDDEN,
            self::CONFIG_STREAM_ALLOW,
            self::CONFIG_STREAM_ENFORCE,
        ])) {
            throw new Exception(sprintf(
                'Stream configuration must be either %s (forbidden), %s (allowed) or %s (enforced)',
                self::CONFIG_STREAM_FORBIDDEN,
                self::CONFIG_STREAM_ALLOW,
                self::CONFIG_STREAM_ENFORCE
            ), self::$FATAL_ERROR);
        }
    }

    private function check_no_errors_occurred()
    {
        $no_errors = empty($this->instrument_errors) && empty($this->pattern_errors) && empty($this->track_errors);
        if ($no_errors) {
            return true;
        }

        $show_max_pattern_errors = 10;
        $number_of_pattern_errors = count($this->pattern_errors);

        $all_errors = $this->instrument_errors;
        $all_errors = array_merge($all_errors, $this->track_errors);

        $all_errors = array_merge($all_errors, array_slice($this->pattern_errors, 0, $show_max_pattern_errors));

        if ($number_of_pattern_errors > $show_max_pattern_errors) {
            $all_errors[] = sprintf('... (%s more pattern errors)', $number_of_pattern_errors - $show_max_pattern_errors);
        }

        throw new Exception(implode(PHP_EOL, $all_errors));
    }

    private function information()
    {
        $str = 'Module name: ' . $this->mod_name . PHP_EOL;
        $str .= 'Number of tracks: ' . $this->no_of_tracks . PHP_EOL;
        $str .= 'Number of patterns: ' . $this->max_pat_cnt . PHP_EOL;
        $str .= 'Number of pattern positions: ' . count($this->pattern_order) . PHP_EOL;
        $str .= PHP_EOL;
        $str .= $this->instrument_osc_requirements();
        $str .= PHP_EOL;
        $str .= 'NTP Instruments:' . PHP_EOL;
        foreach ($this->ntp_instruments as $ntp_instrument) {

            $str .= 'number=' . str_pad($ntp_instrument->number, 3, ' ', STR_PAD_LEFT);
            $str .= ', modnumber=' . str_pad($ntp_instrument->number_original, 3, ' ', STR_PAD_LEFT);
            $str .= ', type=%' . str_pad(decbin($ntp_instrument->get_type_for_ntp()), 4, '0', STR_PAD_LEFT);
            $str .= ', streamed=' . ($ntp_instrument->is_streamed() ? 'yes' : 'no ');
            $str .= ', length=' . str_pad($ntp_instrument->length, 5, ' ', STR_PAD_LEFT);
            $str .= ', volume=' . str_pad($ntp_instrument->volume, 2, ' ', STR_PAD_LEFT);
            $str .= ', finetune=' . str_pad($ntp_instrument->finetune, 2, ' ', STR_PAD_LEFT);
            $str .= ', docptr=$' . str_pad(dechex($this->doc_ptrs[$ntp_instrument->number] * 256), 4, '0', STR_PAD_LEFT);
            $str .= ', docsiz=%' . str_pad(decbin($this->doc_sizes[$ntp_instrument->number]), 6, '0', STR_PAD_LEFT);
            $pos = $this->ntp_instrument_positions[$ntp_instrument->number];
            $str .= ', pos=$' . strtoupper(str_pad(dechex($pos), 6, '0', STR_PAD_LEFT));
            $str .= PHP_EOL;
        }
        $str .= PHP_EOL;
        $str .= 'modnumber - Instrument number in the original module' . PHP_EOL;
        $str .= 'type - Bitmap field, see file format in source code archive' . PHP_EOL;
        $str .= 'docptr - Pointer to which page in sound ram the instrument is placed (streamed samples have a 0 here).' . PHP_EOL;
        $str .= 'docsiz - Bitmap field, see Apple IIGS hardware reference.' . PHP_EOL;
        $str .= 'pos - Position in NTP file where the instrument is stored at.' . PHP_EOL;
        $str .= PHP_EOL;

        $str .= 'DOC ram usage:';
        $doc_pages = $this->instrumentcontainer->get_doc_ram_usage();
        for ($i = 0; $i < 256; $i++) {
            if (fmod($i, 16) == 0) {
                $str .= PHP_EOL;
                $str .= '$' . str_pad(dechex($i * 256), 4, '0', STR_PAD_LEFT) . ' ';
            }
            $str .= $doc_pages[$i];
        }
        $str .= PHP_EOL;
        $str .= '0=free, 1=instrument, 2=stream buffer, 3=interrupt timer' . PHP_EOL;
        $str .= PHP_EOL;

        $str .= 'Oscillator usage:' . PHP_EOL;
        foreach ($this->oscillator_usage as $osc => $value) {
            $str .= 'Oscillator ' . str_pad($osc, 2, '0', STR_PAD_LEFT) . ': ';
            if ($value === 0) {
                $str .= 'free';
            } elseif ($value === -1) {
                $str .= 'Music timer';
            } else {
                $str .= 'Track ' . $value;
            }

            $str .= PHP_EOL;
        }

        return $str;
    }

    public function convert()
    {
        try {
            /* load mod, check config */
            $this->load_mod();
            $this->check_config();

            /* read basic information from mod */
            $this->detect_mod_and_number_of_tracks();
            $this->detect_number_of_patterns();
            $this->calc_pattern_length();
            $this->get_pattern_order();

            /* check patterns for issues */
            $this->analyze_patterns();

            /* convert instruments */
            $this->extract_instruments();
            $this->optimize_instruments();

            /* convert patterns */
            $this->convert_patterns();

            /* assign oscillators for each track */
            $this->determine_osc_per_track();

            /* save, if there are no errors */
            if ($this->check_no_errors_occurred()) {
                $this->save_ntp();
            }

            $result = ['success' => true, 'message' => $this->information() . PHP_EOL];
        } catch (Exception $ex) {
            $result = ['success' => false, 'message' => $ex->getMessage() . PHP_EOL];
        }

        return $result;
    }
}
